package ap;

import ap.proof.{ConstraintSimplifier, ProofTree, ModelSearchProver, Timeout}
import ap.terfor.conjunctions.{Conjunction, Quantifier}
import ap.parameters.{GoalSettings, PreprocessingSettings}

object IntelliFileProver {
  
  abstract class Result
  
  case class ProofResult extends Result
  case class Proof(tree : ProofTree, model : Conjunction) extends ProofResult
  case class NoProof(unsatisfiableTree : ProofTree) extends ProofResult
  case class TimeoutProof(unfinishedTree : ProofTree) extends ProofResult
  
  abstract class CounterModelResult extends Result
  case class CounterModel(counterModel : Conjunction) extends CounterModelResult
  case object NoCounterModel extends CounterModelResult
  case object TimeoutModel extends CounterModelResult
  
}

/**
 * A prover that decides, depending on the kind of the problem, whether it
 * should try to construct a proof tree or just search for counterexamples
 */
class IntelliFileProver(reader : java.io.Reader,
                        simplifier : ConstraintSimplifier,
                        // a timeout in milliseconds
                        timeout : Int,
                        output : Boolean,
                        userDefStoppingCond : => Boolean,
                        preprocSettings : PreprocessingSettings,
                        goalSettings : GoalSettings)
      extends AbstractFileProver(reader, output, timeout,
                                 userDefStoppingCond, preprocSettings,
                                 goalSettings) {

  import IntelliFileProver._
        
  lazy val proofResult : ProofResult =
    Timeout.catchTimeout[ProofResult] {
      val tree = constructProofTree(simplifier)
      val model = findModel(tree.closingConstraint)
      if (model.isFalse)
        NoProof(tree)
      else
        Proof(tree, model)
    } {
      // we assume that a partial proof tree is given back
      case x : ProofTree => TimeoutProof(x)
    }
        
  lazy val proofTree : ProofTree = proofResult match {
    case TimeoutProof(t) => t
    case Proof(t, _) => t
    case NoProof(t) => t
  } 

  lazy val counterModelResult : CounterModelResult =
    Timeout.catchTimeout[CounterModelResult] { 
      val model = findCounterModel
      if (model.isFalse)
        NoCounterModel
      else
        CounterModel(model)
    } {
      case _ => TimeoutModel
    }

  val result : IntelliFileProver.Result = {
    if ((formula.constants subsetOf eliminatedConstants) &&
        (Conjunction.collectQuantifiers(formula) subsetOf Set(Quantifier.ALL))) {
      // try to find a countermodel
      counterModelResult
    } else {
      // try to construct a proof
      proofResult
    }
  }
  
}
