package ap.parser;

import ap.terfor.{VariableTerm, ConstantTerm, TermOrder}
import ap.util.FilterIt

object Environment {
  
  abstract class SymKind
  case object Universal extends SymKind
  case object Existential extends SymKind

  abstract class DeclaredSym
  case class Constant(c : ConstantTerm, k : SymKind) extends DeclaredSym
  case class Variable(index : Int) extends DeclaredSym
  case class Predicate(pred : ap.terfor.preds.Predicate) extends DeclaredSym
  case class Function(fun : IFunction) extends DeclaredSym

  class EnvironmentException(msg : String) extends Exception(msg)

}

class Environment {

  import Environment._
  
  /** The declared symbols */
  private val signature = new scala.collection.mutable.HashMap[String, DeclaredSym]
  
  /** The variables bound at the present point */
  private val context = new scala.collection.mutable.ArrayBuffer[String]
  
  /** A <code>TermOrder</code> containing all declared constants */
  private var orderVar = TermOrder.EMPTY
  
  def order = orderVar
  
  def lookupSym(name : String) : DeclaredSym = (context lastIndexOf name) match {
    case -1 => (signature get name) match {
      case Some(t) => t
      case None => throw new EnvironmentException("Symbol " + name + " not declared")
    }
    case index => Variable(context.size - index - 1)
  }
  
  def addConstant(c : ConstantTerm, kind : SymKind) : Unit = {
    addSym(c.name, Constant(c, kind))
    // if the constant is universal, make it maximal; otherwise,
    // make it just bigger than all existential constants
    orderVar = kind match {
      case Universal => orderVar.extend(c, Set.empty)
      case Existential => orderVar.extend(c, universalConstants)
    }
  }
 
  def addPredicate(pred : ap.terfor.preds.Predicate) : Unit = {
    addSym(pred.name, Predicate(pred))
    orderVar = orderVar extend pred
  }
  
  def addFunction(fun : IFunction) : Unit =
    addSym(fun.name, Function(fun))
  
  def universalConstants : Set[ConstantTerm] = {
    val predicate : DeclaredSym => boolean = {
      case Constant(_, Universal) => true
      case _ => false
    }
    Set.empty ++ (for (Constant(c, _) <- FilterIt(signature.values, predicate))
                  yield c)
  }
  
  private def addSym(name : String, t : DeclaredSym) : Unit =
    if (signature contains name)
      throw new EnvironmentException("Symbol " + name + " is already declared")
    else
      signature += (name -> t)
  
  def pushVar(name : String) : Unit = (context += name)

  def popVar : Unit =
    if (context isEmpty)
      throw new EnvironmentException("Trying to pop a non-existing variable")
    else
      context reduceToSize (context.size - 1)
  
}
