package ap.terfor.conjunctions;

import ap.terfor.arithconj.ArithConj
import ap.terfor.preds.{PredConj, Atom}
import ap.terfor.equations.EquationConj
import ap.util.{Debug, Logic, Seqs}

import scala.collection.mutable.ArrayBuffer
import scala.util.Sorting

object ClauseMatcher {
  
  private val AC = Debug.AC_CLAUSE_MATCHER
  
}

/**
 * Class for matching certain predicate literals of a clause (an existentially
 * quantified conjunction) on facts.
 */
class ClauseMatcher(// facts that the clause has already been matched on
                    oldFacts : PredConj,
                    // new facts that have not been available previously
                    newFacts : PredConj,
                    contextReducer : ReduceWithConjunction,
                    // predicate to distinguish the relevant matches
                    // (e.g., to filter out shielded formulae)
                    isIrrelevantMatch : (Conjunction) => Boolean,
                    order : TermOrder) {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(ClauseMatcher.AC,
                   (oldFacts isSortedBy order) && (newFacts isSortedBy order))
  //////////////////////////////////////////////////////////////////////////////
  
  /**
   * For a certain atom, determine whether the positive occurrences or the
   * negative occurrences are to be resolved.
   *
   * TODO: generalise this so that for some predicates the positive occurrences,
   * for some predicates the negative occurrences are resolved
   */
  private def isPositivelyMatched(a : Atom) : Boolean = true

  // TODO: one should perform real subsumption tests here
  val resolvedClauses = new scala.collection.mutable.HashSet[Conjunction]
  
  /**
   * Match the predicate literals of an existentially quantified conjunction
   * on the given facts and store the resolvants in <code>resolvedClauses</code> 
   */
  def matchClause(clause : Conjunction) : Unit = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ClauseMatcher.AC,
                    (clause isSortedBy order) &&
                    Logic.forall(for (q <- clause.quans.elements)
                                 yield (q == Quantifier.EX)))
    ////////////////////////////////////////////////////////////////////////////
    
    val (matchedLits, remainingLits) = determineMatchedLits(clause.predConj)
    
    val res = new ArrayBuffer[Conjunction]
    for (conds <- matchLiterals(clause.arithConj, matchedLits)) {
      val newConj = Conjunction(clause.quans, conds, remainingLits,
                                clause.negatedConjs, order)
      val reducedConj = contextReducer(newConj)
      if (!reducedConj.isFalse) resolvedClauses += reducedConj
    }
  }
  
  /**
   * Given a conjunction of atoms, determine which atoms are to be matched on
   * facts. The 2nd component of the result are the remaining literals.
   */
  private def determineMatchedLits(conj : PredConj) : (Seq[Atom], PredConj) = {
    val (posMatched, posRemaining) =
      Seqs.split(conj.positiveLits, (a:Atom) => isPositivelyMatched(a))
    val (negMatched, negRemaining) =
      Seqs.split(conj.negativeLits, (a:Atom) => !isPositivelyMatched(a))

    (posMatched ++ negMatched,
     conj.updateLitsSubset(posRemaining, negRemaining, order))
  }
  
  /**
   * Match all <code>atoms</code> on the given facts. The resulting equations
   * are conjunctively added to <code>initialEqs</code>.
   */
  private def matchLiterals(initialEqs : ArithConj,
                            atoms : Seq[Atom]) : Seq[ArithConj] = {
    val (wNewMatches, woNewMatches) =
      Seqs.split(atoms, (a:Atom) => !newMatches(a).isEmpty)
    
    if (wNewMatches.isEmpty) {
      List()
    } else {
      val wnmAtoms = sortAtomsToBeMatched(wNewMatches)
      val wonmAtoms = sortAtomsToBeMatched(woNewMatches)
      
      combineMatches(initialEqs, wnmAtoms, false)
      for (eqs <- clearMatchResults) combineMatches(eqs, wonmAtoms, true)
        
      clearMatchResults
    }
  }
  
  //////////////////////////////////////////////////////////////////////////////
  
  private def sortAtomsToBeMatched(atoms : RandomAccessSeq[Atom]) : List[Atom] = {
    val atomsArray = atoms.toArray
    Sorting.stableSort(atomsArray,
                       (a1 : Atom, a2 : Atom) => matchNum(a1) < matchNum(a2))
    atomsArray.toList
  }
  
  /**
   * Recursively match <code>atoms</code> on the given facts in all possible
   * ways. <code>newMatchesIncluded</code> tells whether one of the
   * <code>newFacts</code> has already been included in the matching, i.e.,
   * for all following atoms all facts can be considered (also
   * <code>oldFacts</code>)
   */
  private def combineMatches(currentEqs : ArithConj,
                             atoms : List[Atom],
                             newMatchIncluded : Boolean) : Unit = atoms match {
    case List() => addMatchResult(currentEqs)
    case a :: otherAtoms => {
      def handleNewEqs(newEqs : EquationConj, nmi : Boolean) : Unit =
        simplify(ArithConj.conj(Array(currentEqs, newEqs), order),
                 (ac:ArithConj) => combineMatches(ac, otherAtoms, nmi))
      
      newMatches(a).foreach(handleNewEqs(_, true))
      if (newMatchIncluded || !otherAtoms.isEmpty)
        oldMatches(a).foreach(handleNewEqs(_, newMatchIncluded))
    }
  }
  
  private val matchResults = new ArrayBuffer[ArithConj]
  
  private def addMatchResult(eqs : ArithConj) : Unit =
    matchResults += eqs
  
  private def clearMatchResults : Seq[ArithConj] = {
    val res = matchResults.toArray
    matchResults.clear
    res
  }

  //////////////////////////////////////////////////////////////////////////////
  
  private def simplify(eqs : EquationConj, cont : (EquationConj) => Unit) : Unit = {
    val reduced = contextReducer(eqs)
    if (!reduced.isFalse
        && !isIrrelevantMatch(Conjunction.conj(reduced, order)))
      cont(reduced)
  }

  private def simplify(ac : ArithConj, cont : (ArithConj) => Unit) : Unit = {
    val reduced = contextReducer(ac)
    if (!reduced.isFalse
        && !isIrrelevantMatch(Conjunction.conj(reduced, order)))
      cont(reduced)
  }

  //////////////////////////////////////////////////////////////////////////////
  
  private def matchNum(a : Atom) : Int = oldMatches(a).size + newMatches(a).size
  
  private val oldMatchesCache =
    new scala.collection.mutable.HashMap[Atom, Seq[EquationConj]]
  private val newMatchesCache =
    new scala.collection.mutable.HashMap[Atom, Seq[EquationConj]]
  
  private def oldMatches(a : Atom) =
    oldMatchesCache.getOrElseUpdate(a, matches(a, oldFacts))

  private def newMatches(a : Atom) =
    newMatchesCache.getOrElseUpdate(a, matches(a, newFacts))

  private def matches(a : Atom, facts : PredConj) : Seq[EquationConj] = {
    val res = new ArrayBuffer[EquationConj]
    val lits = if (isPositivelyMatched(a))
                 facts positiveLitsWithPred a.pred
               else
                 facts negativeLitsWithPred a.pred
    // TODO: make this more efficient
    val adder = res += _
    for (atom2 <- lits) simplify(a.unify(atom2, order), adder)
    res
  }
    
}
